@@ -693,18 +693,6 @@ polygonsStoreClips(CompScreen * s, CompWindow * w,
     if (!pset)
 	return;
 
-    // only draw windows on current viewport
-    if (w->attrib.x > s->width || w->attrib.x + w->width < 0 ||
-	w->attrib.y > s->height || w->attrib.y + w->height < 0 ||
-	(aw->lastKnownCoords.x != NOT_INITIALIZED &&
-	 (aw->lastKnownCoords.x != w->attrib.x ||
-	  aw->lastKnownCoords.y != w->attrib.y)))
-    {
-	return;
-	// since this is not the viewport the window was drawn
-	// just before animation started
-    }
-
     Bool dontStoreClips = TRUE;
 
     // If this clip doesn't match the corresponding stored clip,
@@ -883,7 +871,7 @@ static void
 getPerspectiveCorrectionMat (CompWindow *w,
 			     PolygonObject *p,
 			     GLfloat *mat,
-			     float *matf)
+			     CompTransform *matf)
 {
     CompScreen *s = w->screen;
     ANIM_SCREEN (s);
@@ -925,7 +913,47 @@ getPerspectiveCorrectionMat (CompWindow *w,
 	     0,1,0,0,
 	     skewx,skewy,1,0,
 	     0,0,0,1};
-	memcpy (matf, skewMat, 16 * sizeof (float));
+	memcpy (matf->m, skewMat, 16 * sizeof (float));
+    }
+}
+
+static void
+prepareDrawingForAttrib (CompScreen *s,
+			 FragmentAttrib *attrib)
+{
+    if (s->canDoSaturated && attrib->saturation != COLOR)
+    {
+	GLfloat constant[4];
+
+	if (s->canDoSlightlySaturated && attrib->saturation > 0)
+	{
+	    constant[3] = attrib->opacity / 65535.0f;
+	    constant[0] = constant[1] = constant[2] = constant[3] *
+		attrib->brightness / 65535.0f;
+
+	    glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, constant);
+	}
+	else
+	{
+	    constant[3] = attrib->opacity / 65535.0f;
+	    constant[0] = constant[1] = constant[2] = constant[3] *
+		attrib->brightness / 65535.0f;
+
+	    constant[0] = 0.5f + 0.5f * RED_SATURATION_WEIGHT   * constant[0];
+	    constant[1] = 0.5f + 0.5f * GREEN_SATURATION_WEIGHT * constant[1];
+	    constant[2] = 0.5f + 0.5f * BLUE_SATURATION_WEIGHT  * constant[2];
+
+	    glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, constant);
+	}
+    }
+    else
+    {
+	attrib->brightness *= 0.76;
+
+	GLushort color = (attrib->opacity * attrib->brightness) >> 16;
+
+	screenTexEnvMode (s, GL_MODULATE);
+	glColor4us (color, color, color, attrib->opacity);
     }
 }
 
@@ -933,12 +961,9 @@ void polygonsDrawCustomGeometry(CompScreen * s, CompWindow * w)
 {
     ANIM_WINDOW(w);
 
-    if (// only draw windows on current viewport
-	w->attrib.x > s->width || w->attrib.x + w->width < 0 ||
-	w->attrib.y > s->height || w->attrib.y + w->height < 0 ||
-	(aw->lastKnownCoords.x != NOT_INITIALIZED &&
-	 (aw->lastKnownCoords.x != w->attrib.x ||
-	  aw->lastKnownCoords.y != w->attrib.y)))
+    // draw windows only on current viewport unless it's on all viewports
+    if ((s->windowOffsetX != 0 || s->windowOffsetY != 0) &&
+	!windowOnAllViewports (w))
     {
 	return;
 	// since this is not the viewport the window was drawn
@@ -1003,11 +1028,7 @@ void polygonsDrawCustomGeometry(CompScreen * s, CompWindow * w)
     }
     glPushMatrix();
 
-    // Store old blend values
-    GLint blendSrcWas, blendDstWas;
-
-    glGetIntegerv(GL_BLEND_SRC, &blendSrcWas);
-    glGetIntegerv(GL_BLEND_DST, &blendDstWas);
+    // Store old blend
     GLboolean blendWas = glIsEnabled(GL_BLEND);
 
     glPushAttrib(GL_STENCIL_BUFFER_BIT);
@@ -1036,7 +1057,6 @@ void polygonsDrawCustomGeometry(CompScreen * s, CompWindow * w)
 
     glGetFloatv(GL_CURRENT_COLOR, oldColor);
 
-
     // Determine where we are called from in paint.c's drawWindowTexture
     // to find out how we should change the opacity
     GLint prevActiveTexture = GL_TEXTURE0_ARB;
@@ -1058,26 +1078,16 @@ void polygonsDrawCustomGeometry(CompScreen * s, CompWindow * w)
 	    prevActiveTexture = GL_TEXTURE1_ARB;
     }
 
-    float brightness = aw->curPaintAttrib.brightness / 65535.0;
     float opacity = aw->curPaintAttrib.opacity / 65535.0;
 
     float newOpacity = opacity;
     float fadePassedBy;
 
     if (!blendWas)				// if translucency is not already turned on in paint.c
-    {
 	glEnable(GL_BLEND);
-    }
+
     if (saturationFull)
-    {
 	screenTexEnvMode(w->screen, GL_MODULATE);
-    }
-    else if (prevActiveTexture == GL_TEXTURE2_ARB)
-    {
-	w->screen->activeTexture(prevActiveTexture + 1);
-	enableTexture(w->screen, aw->curTexture, aw->curTextureFilter);
-    }
-    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 
     // if fade-out duration is not specified per polygon
     if (pset->allFadeDuration > -1.0f)
@@ -1221,54 +1231,9 @@ void polygonsDrawCustomGeometry(CompScreen * s, CompWindow * w)
 			(forwardProgress / pset->backAndSidesFadeDur);
 		}
 
-		if (saturationFull)
-		    glColor4f(brightness, brightness, brightness,
-			      newOpacityPolygon2);
-		else if (prevActiveTexture == GL_TEXTURE1_ARB)
-		{
-		    // From paint.c
-
-		    GLfloat constant2[4] =
-			{ 0.5f +
-			  0.5f * RED_SATURATION_WEIGHT * brightness,
-			  0.5f + 0.5f * GREEN_SATURATION_WEIGHT * brightness,
-			  0.5f + 0.5f * BLUE_SATURATION_WEIGHT * brightness,
-			  newOpacityPolygon2
-			};
-
-		    glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR,
-			       constant2);
-		}
-		else			//if (prevActiveTexture >= GL_TEXTURE2_ARB)
-		{
-		    GLfloat constant2[4] = { brightness,
-					     brightness,
-					     brightness,
-					     newOpacityPolygon2
-		    };
-
-		    // From paint.c
-
-		    glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR,
-			       constant2);
-
-		    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,
-			      GL_COMBINE);
-
-		    glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
-		    glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
-		    glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_CONSTANT);
-		    glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
-		    glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
-
-		    glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_ALPHA, GL_MODULATE);
-		    glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA, GL_PREVIOUS);
-		    glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA, GL_CONSTANT);
-		    glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA,
-			      GL_SRC_ALPHA);
-		    glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA,
-			      GL_SRC_ALPHA);
-		}
+		FragmentAttrib attrib = aw->curPaintAttrib;
+		attrib.opacity = newOpacityPolygon2 * OPAQUE;
+		prepareDrawingForAttrib (s, &attrib);
 
 		//glEnableClientState(GL_NORMAL_ARRAY);
 		//glEnable(GL_NORMALIZE);
@@ -1311,41 +1276,15 @@ void polygonsDrawCustomGeometry(CompScreen * s, CompWindow * w)
 		    glDrawElements(GL_QUAD_STRIP, 2 * (p->nSides + 1),
 				   GL_UNSIGNED_SHORT, p->sideIndices);
 
+		// if opacity was changed just above
 		if (fadeBackAndSides)
-		    // if opacity was changed just above
 		{
 		    // Go back to normal opacity for front face
-
-		    if (saturationFull)
-			glColor4f(brightness, brightness, brightness,
-				  newOpacityPolygon);
-		    else if (prevActiveTexture == GL_TEXTURE1_ARB)
-		    {
-			GLfloat constant[4] =
-			    { 0.5f +
-			      0.5f * RED_SATURATION_WEIGHT * brightness,
-			      0.5f + 0.5f * GREEN_SATURATION_WEIGHT *
-			      brightness,
-			      0.5f + 0.5f * BLUE_SATURATION_WEIGHT * brightness,
-			      newOpacityPolygon
-			    };
-
-			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR,
-				   constant);
-		    }
-		    else
-		    {
-			GLfloat constant[4] = { brightness,
-						brightness,
-						brightness,
-						newOpacityPolygon
-			};
-			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR,
-				   constant);
-		    }
+		    attrib = aw->curPaintAttrib;
+		    attrib.opacity = newOpacityPolygon * OPAQUE;
+		    prepareDrawingForAttrib (s, &attrib);
 		}
 		// Draw front face
-
 		glNormal3f(p->normals[0], p->normals[1], p->normals[2]);
 		glDrawArrays(GL_POLYGON, 0, p->nSides);
 
@@ -1368,20 +1307,13 @@ void polygonsDrawCustomGeometry(CompScreen * s, CompWindow * w)
 	glPopAttrib();
 	glPopAttrib();
     }
-
     // Restore texture stuff
     if (saturationFull)
 	screenTexEnvMode(w->screen, GL_REPLACE);
-    else if (prevActiveTexture == GL_TEXTURE2_ARB)
-    {
-	disableTexture(w->screen, aw->curTexture);
-	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
-	w->screen->activeTexture(prevActiveTexture);
-    }
-    // Restore blend values
+
+    // Restore blend
     if (!blendWas)
 	glDisable(GL_BLEND);
-    glBlendFunc(blendSrcWas, blendDstWas);
 
     glPopMatrix();
 
@@ -1465,10 +1397,10 @@ polygonsDeceleratingAnimStepPolygon(CompWindow * w,
     p->rotAngle = moveProgress * p->finalRotAng + p->rotAngleStart;
 }
 
-Bool polygonsAnimStep(CompScreen * s, CompWindow * w, float time)
+void
+polygonsAnimStep (CompScreen *s, CompWindow *w, float time)
 {
-    if (!defaultAnimStep(s, w, time))
-	return FALSE;
+    defaultAnimStep (s, w, time);
 
     ANIM_WINDOW(w);
 
@@ -1490,7 +1422,6 @@ Bool polygonsAnimStep(CompScreen * s, CompWindow * w, float time)
     else
 	compLogMessage (s->display, "animation", CompLogLevelDebug,
 			"%s: pset null at line %d\n",__FILE__,  __LINE__);
-    return TRUE;
 }
 
 void
@@ -1507,7 +1438,7 @@ polygonsUpdateBB (CompOutput *output,
     CompTransform wTransform;
     CompTransform wTransform2;
 
-    resetToIdentity (&wTransform2);
+    matrixGetIdentity (&wTransform2);
     prepareTransform (s, output, &wTransform, &wTransform2);
 
     GLdouble dModel[16];
@@ -1527,11 +1458,11 @@ polygonsUpdateBB (CompOutput *output,
     PolygonObject *p = aw->polygonSet->polygons;
     CompTransform *modelViewTransform = &wTransform;
 
-    float skewMat[16];
+    CompTransform skewMat;
     if (pset->correctPerspective == CorrectPerspectiveWindow)
     {
-	getPerspectiveCorrectionMat (w, NULL, NULL, skewMat);
-	matmul4 (wTransform2.m, wTransform.m, skewMat);
+	getPerspectiveCorrectionMat (w, NULL, NULL, &skewMat);
+	matrixMultiply (&wTransform2, &wTransform, &skewMat);
     }
     if (pset->correctPerspective == CorrectPerspectiveWindow ||
 	pset->correctPerspective == CorrectPerspectivePolygon)
@@ -1541,8 +1472,8 @@ polygonsUpdateBB (CompOutput *output,
     {
 	if (pset->correctPerspective == CorrectPerspectivePolygon)
 	{
-	    getPerspectiveCorrectionMat (w, p, NULL, skewMat);
-	    matmul4 (wTransform2.m, wTransform.m, skewMat);
+	    getPerspectiveCorrectionMat (w, p, NULL, &skewMat);
+	    matrixMultiply (&wTransform2, &wTransform, &skewMat);
 	}
 
 	// if modelViewTransform == wTransform2, then
@@ -1554,9 +1485,25 @@ polygonsUpdateBB (CompOutput *output,
 	Point3d center = p->centerPos;
 	float radius = p->boundSphereRadius + 2;
 
+	// Take rotation axis offset into consideration and
+	// properly enclose polygon in the bounding cube
+	// whatever the rotation angle is:
+
+	// Add rotation axis offset to center (rotated) polygon correctly
+	// within bounding cube
+	center.x += p->rotAxisOffset.x;
+	center.y += p->rotAxisOffset.y;
+	center.z += p->rotAxisOffset.z / s->width;
+
+	// Add rotation axis offset to radius to enlarge the bounding cube
+	radius += MAX (MAX (fabs(p->rotAxisOffset.x),
+			    fabs(p->rotAxisOffset.y)),
+		       fabs(p->rotAxisOffset.z));
+
 	float zradius = radius / s->width;
+
 #define N_POINTS 8
-	// Corners of almost-bounding rect. prism
+	// Corners of bounding cube
 	Point3d cubeCorners[N_POINTS] =
 	    {{center.x - radius, center.y - radius, center.z + zradius},
 	     {center.x - radius, center.y + radius, center.z + zradius},
