@@ -32,9 +32,9 @@
 #include <string.h>
 #include <X11/Xatom.h>
 
-#include <compiz.h>
-#include <scale.h>
-#include <text.h>
+#include <compiz-core.h>
+#include <compiz-scale.h>
+#include <compiz-text.h>
 
 #include "scaleaddon_options.h"
 
@@ -52,7 +52,10 @@ typedef struct _ScaleAddonDisplay {
     HandleEventProc       handleEvent;
     HandleCompizEventProc handleCompizEvent;
 
-    Window lastHoveredWindow;
+    Bool textAvailable;
+
+    Window highlightedWindow;
+    Window lastHighlightedWindow;
 } ScaleAddonDisplay;
 
 typedef struct _ScaleAddonScreen {
@@ -60,12 +63,10 @@ typedef struct _ScaleAddonScreen {
 
     ScaleLayoutSlotsAndAssignWindowsProc layoutSlotsAndAssignWindows;
     ScalePaintDecorationProc		 scalePaintDecoration;
+    ScaleSelectWindowProc                selectWindow;
+    DonePaintScreenProc                  donePaintScreen;
 
-    Pixmap      textPixmap;
-    CompTexture textTexture;
-
-    int textWidth;
-    int textHeight;
+    int lastState;
 
     float scale;
 } ScaleAddonScreen;
@@ -73,112 +74,151 @@ typedef struct _ScaleAddonScreen {
 typedef struct _ScaleAddonWindow {
     ScaleSlot origSlot;
 
+    Pixmap      textPixmap;
+    CompTexture textTexture;
+
+    int textWidth;
+    int textHeight;
+
     Bool rescaled;
 
     CompWindow *oldAbove;
 } ScaleAddonWindow;
 
 #define GET_ADDON_DISPLAY(d)				          \
-    ((ScaleAddonDisplay *) (d)->privates[displayPrivateIndex].ptr)
+    ((ScaleAddonDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define ADDON_DISPLAY(d)		          \
     ScaleAddonDisplay *ad = GET_ADDON_DISPLAY (d)
 
 #define GET_ADDON_SCREEN(s, ad)				              \
-    ((ScaleAddonScreen *) (s)->privates[(ad)->screenPrivateIndex].ptr)
+    ((ScaleAddonScreen *) (s)->base.privates[(ad)->screenPrivateIndex].ptr)
 
 #define ADDON_SCREEN(s)						               \
     ScaleAddonScreen *as = GET_ADDON_SCREEN (s, GET_ADDON_DISPLAY (s->display))
 
 #define GET_ADDON_WINDOW(w, as)				              \
-    ((ScaleAddonWindow *) (w)->privates[(as)->windowPrivateIndex].ptr)
+    ((ScaleAddonWindow *) (w)->base.privates[(as)->windowPrivateIndex].ptr)
 
-#define ADDON_WINDOW(w)						                 \
-    ScaleAddonWindow *aw = GET_ADDON_WINDOW (w,                                  \
-	    GET_ADDON_SCREEN (w->screen, GET_ADDON_DISPLAY (w->screen->display)))
+#define ADDON_WINDOW(w)						   \
+    ScaleAddonWindow *aw = GET_ADDON_WINDOW (w,                    \
+	                   GET_ADDON_SCREEN (w->screen,            \
+			   GET_ADDON_DISPLAY (w->screen->display)))
 
 
 static void
-scaleaddonFreeWindowTitle (CompScreen *s)
+scaleaddonFreeWindowTitle (CompWindow *w)
 {
-    ADDON_SCREEN (s);
+    CompScreen *s = w->screen;
 
-    if (!as->textPixmap)
+    ADDON_WINDOW (w);
+
+    if (!aw->textPixmap)
 	return;
 
-    releasePixmapFromTexture(s, &as->textTexture);
-    initTexture (s, &as->textTexture);
-    XFreePixmap (s->display->display, as->textPixmap);
-    as->textPixmap = None;
+    releasePixmapFromTexture (s, &aw->textTexture);
+    initTexture (s, &aw->textTexture);
+    XFreePixmap (s->display->display, aw->textPixmap);
+    aw->textPixmap = None;
 }
 
 static void
 scaleaddonRenderWindowTitle (CompWindow *w)
 {
-    CompTextAttrib tA;
-    float          scale;
-    int            stride;
-    void*          data;
+    CompTextAttrib            tA;
+    float                     scale;
+    int                       stride;
+    void                      *data;
+    CompScreen                *s = w->screen;
+    ScaleaddonWindowTitleEnum winTitleMode;
 
-    ADDON_SCREEN (w->screen);
+    ADDON_DISPLAY (s->display);
+    SCALE_SCREEN (s);
     SCALE_WINDOW (w);
+    ADDON_WINDOW (w);
 
-    scaleaddonFreeWindowTitle(w->screen);
+    scaleaddonFreeWindowTitle (w);
 
-    if (!scaleaddonGetWindowTitle (w->screen))
+    if (!ad->textAvailable)
 	return;
 
+    if (!sw->slot)
+	return;
+
+    winTitleMode = scaleaddonGetWindowTitle (s);
+    if (winTitleMode == WindowTitleNoDisplay)
+	return;
+
+    if (winTitleMode == WindowTitleHighlightedWindowOnly &&
+	ad->highlightedWindow != w->id)
+    {
+	return;
+    }
+
     scale = sw->slot ? sw->slot->scale : sw->scale;
-    tA.maxwidth = (w->attrib.width * scale) -
-	          (2 * scaleaddonGetBorderSize (w->screen));
-    tA.maxheight = (w->attrib.height * scale) -
-     	           (2 * scaleaddonGetBorderSize (w->screen));
-    tA.screen = w->screen;
-    tA.size = scaleaddonGetTitleSize (w->screen);
-    tA.color[0] = scaleaddonGetFontColorRed (w->screen);
-    tA.color[1] = scaleaddonGetFontColorGreen (w->screen);
-    tA.color[2] = scaleaddonGetFontColorBlue (w->screen);
-    tA.color[3] = scaleaddonGetFontColorAlpha (w->screen);
-    tA.style = (scaleaddonGetTitleBold (w->screen)) ?
+    tA.maxWidth = (w->attrib.width * scale) - (2 * scaleaddonGetBorderSize (s));
+    tA.maxHeight = (w->attrib.height * scale) - 
+	           (2 * scaleaddonGetBorderSize (s));
+    tA.screen = s;
+    tA.size = scaleaddonGetTitleSize (s);
+    tA.color[0] = scaleaddonGetFontColorRed (s);
+    tA.color[1] = scaleaddonGetFontColorGreen (s);
+    tA.color[2] = scaleaddonGetFontColorBlue (s);
+    tA.color[3] = scaleaddonGetFontColorAlpha (s);
+    tA.style = (scaleaddonGetTitleBold (s)) ?
 	       TEXT_STYLE_BOLD : TEXT_STYLE_NORMAL;
     tA.family = "Sans";
     tA.ellipsize = TRUE;
 
-    tA.renderMode = TextRenderWindowTitle;
+    if (ss->type == ScaleTypeAll)
+	tA.renderMode = TextRenderWindowTitleWithViewport;
+    else
+	tA.renderMode = TextRenderWindowTitle;
+
     tA.data = (void*)w->id;
 
-    if ((*w->screen->display->fileToImage) (w->screen->display, TEXT_ID, (char *)&tA,
-					    &as->textWidth, &as->textHeight, &stride, &data))
+    if ((*s->display->fileToImage) (s->display, TEXT_ID, (char *)&tA,
+				    &aw->textWidth, &aw->textHeight,
+				    &stride, &data))
     {
-	as->textPixmap = (Pixmap)data;
-	bindPixmapToTexture (w->screen, &as->textTexture, as->textPixmap,
-			     as->textWidth, as->textHeight, 32);
+	aw->textPixmap = (Pixmap)data;
+	if (!bindPixmapToTexture (s, &aw->textTexture, aw->textPixmap,
+	     			  aw->textWidth, aw->textHeight, 32))
+	{
+	    compLogMessage (s->display, "scaleaddon", CompLogLevelError,
+			    "Bind pixmap to texture failed.\n");
+	    XFreePixmap (s->display->display, aw->textPixmap);
+	    aw->textPixmap = None;
+	    aw->textWidth = 0;
+	    aw->textHeight = 0;
+	}
     }
     else
     {
-	as->textPixmap = None;
-	as->textWidth = 0;
-	as->textHeight = 0;
+	aw->textPixmap = None;
+	aw->textWidth = 0;
+	aw->textHeight = 0;
     }
 }
 
 static void
 scaleaddonDrawWindowTitle (CompWindow *w)
 {
-    ADDON_SCREEN (w->screen);
+    GLboolean  wasBlend;
+    GLint      oldBlendSrc, oldBlendDst;
+    float      x, y, width, height, border;
+    CompScreen *s = w->screen;
+    CompMatrix *m;
+
     SCALE_WINDOW (w);
+    ADDON_WINDOW (w);
 
-    GLboolean wasBlend;
-    GLint     oldBlendSrc, oldBlendDst;
+    width = aw->textWidth;
+    height = aw->textHeight;
+    border = scaleaddonGetBorderSize (s);
 
-    float width = as->textWidth;
-    float height = as->textHeight;
-    float border = scaleaddonGetBorderSize (w->screen);
-
-    float x = sw->tx + w->attrib.x +
-	      ((w->attrib.width * sw->scale) / 2) - (as->textWidth / 2);
-    float y = sw->ty + w->attrib.y +
-	      ((w->attrib.height * sw->scale) / 2) - (as->textHeight / 2);
+    x = sw->tx + w->attrib.x + ((WIN_W (w) * sw->scale) / 2) - (width / 2);
+    y = sw->ty + w->attrib.y + ((WIN_H (w) * sw->scale) / 2) - (height / 2);
 
     x = floor (x);
     y = floor (y);
@@ -192,10 +232,10 @@ scaleaddonDrawWindowTitle (CompWindow *w)
 
     glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
 
-    glColor4us (scaleaddonGetBackColorRed (w->screen),
-		scaleaddonGetBackColorGreen (w->screen),
-		scaleaddonGetBackColorBlue (w->screen),
-		scaleaddonGetBackColorAlpha (w->screen));
+    glColor4us (scaleaddonGetBackColorRed (s),
+		scaleaddonGetBackColorGreen (s),
+		scaleaddonGetBackColorBlue (s),
+		scaleaddonGetBackColorAlpha (s));
 
     glPushMatrix ();
 
@@ -210,7 +250,7 @@ scaleaddonDrawWindowTitle (CompWindow *w)
 #define CORNER(a,b) \
     for (k = a; k < b; k++) \
     {\
-	float rad = k* (3.14159 / 180.0f);\
+	float rad = k * (3.14159 / 180.0f);\
 	glVertex2f (0.0f, 0.0f);\
 	glVertex2f (cos (rad) * border, sin (rad) * border);\
 	glVertex2f (cos ((k - 1) * (3.14159 / 180.0f)) * border, \
@@ -247,24 +287,24 @@ scaleaddonDrawWindowTitle (CompWindow *w)
     glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
     glColor4f (1.0, 1.0, 1.0, 1.0);
 
-    enableTexture (w->screen, &as->textTexture, COMP_TEXTURE_FILTER_GOOD);
+    enableTexture (s, &aw->textTexture, COMP_TEXTURE_FILTER_GOOD);
 
-    CompMatrix *m = &as->textTexture.matrix;
+    m = &aw->textTexture.matrix;
 
     glBegin (GL_QUADS);
 
-    glTexCoord2f (COMP_TEX_COORD_X(m, 0),COMP_TEX_COORD_Y(m ,0));
+    glTexCoord2f (COMP_TEX_COORD_X (m, 0),COMP_TEX_COORD_Y (m ,0));
     glVertex2f (x, y - height);
-    glTexCoord2f (COMP_TEX_COORD_X(m, 0),COMP_TEX_COORD_Y(m, height));
+    glTexCoord2f (COMP_TEX_COORD_X (m, 0),COMP_TEX_COORD_Y (m, height));
     glVertex2f (x, y);
-    glTexCoord2f (COMP_TEX_COORD_X(m, width),COMP_TEX_COORD_Y(m, height));
+    glTexCoord2f (COMP_TEX_COORD_X (m, width),COMP_TEX_COORD_Y (m, height));
     glVertex2f (x + width, y);
-    glTexCoord2f (COMP_TEX_COORD_X(m, width),COMP_TEX_COORD_Y(m, 0));
+    glTexCoord2f (COMP_TEX_COORD_X (m, width),COMP_TEX_COORD_Y (m, 0));
     glVertex2f (x + width, y - height);
 
     glEnd ();
 
-    disableTexture (w->screen, &as->textTexture);
+    disableTexture (s, &aw->textTexture);
     glColor4usv (defaultColor);
 
     if (!wasBlend)
@@ -275,19 +315,21 @@ scaleaddonDrawWindowTitle (CompWindow *w)
 static void
 scaleaddonDrawWindowHighlight (CompWindow *w)
 {
+    GLboolean  wasBlend;
+    GLint      oldBlendSrc, oldBlendDst;
+    float      x, y, width, height;
+    CompScreen *s = w->screen;
+
     SCALE_WINDOW (w);
     ADDON_WINDOW (w);
 
-    GLboolean wasBlend;
-    GLint     oldBlendSrc, oldBlendDst;
-
     if (aw->rescaled)
 	return;
 
-    float x      = sw->tx + w->attrib.x - (w->input.left * sw->scale);
-    float y      = sw->ty + w->attrib.y - (w->input.top * sw->scale);
-    float width  = WIN_W(w) * sw->scale;
-    float height = WIN_H(w) * sw->scale;
+    x      = sw->tx + w->attrib.x - (w->input.left * sw->scale);
+    y      = sw->ty + w->attrib.y - (w->input.top * sw->scale);
+    width  = WIN_W (w) * sw->scale;
+    height = WIN_H (w) * sw->scale;
 
     /* we use a poor replacement for roundf()
      * (available in C99 only) here */
@@ -303,10 +345,10 @@ scaleaddonDrawWindowHighlight (CompWindow *w)
 
     glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 
-    glColor4us (scaleaddonGetHighlightColorRed (w->screen),
-		scaleaddonGetHighlightColorGreen (w->screen),
-		scaleaddonGetHighlightColorBlue (w->screen),
-		scaleaddonGetHighlightColorAlpha (w->screen));
+    glColor4us (scaleaddonGetHighlightColorRed (s),
+		scaleaddonGetHighlightColorGreen (s),
+		scaleaddonGetHighlightColorBlue (s),
+		scaleaddonGetHighlightColorAlpha (s));
 
     glRectf (x, y + height, x + width, y);
 
@@ -318,43 +360,45 @@ scaleaddonDrawWindowHighlight (CompWindow *w)
 }
 
 static void
-scaleaddonCheckHoveredWindow (CompScreen *s)
+scaleaddonCheckWindowHighlight (CompScreen *s)
 {
     CompDisplay *d = s->display;
 
     ADDON_DISPLAY (d);
-    SCALE_DISPLAY (d);
 
-    if (sd->hoveredWindow != ad->lastHoveredWindow)
+    if (ad->highlightedWindow != ad->lastHighlightedWindow)
     {
-	CompWindow *w, *lw;
+	CompWindow *w;
 
-	w = findWindowAtDisplay (d, sd->hoveredWindow);
+	w = findWindowAtDisplay (d, ad->highlightedWindow);
 	if (w)
 	{
 	    scaleaddonRenderWindowTitle (w);
 	    addWindowDamage (w);
 	}
-	else
-	    scaleaddonFreeWindowTitle (s);
 
-	lw = findWindowAtDisplay (d, ad->lastHoveredWindow);
-	if (lw)
-	    addWindowDamage (lw);
+	w = findWindowAtDisplay (d, ad->lastHighlightedWindow);
+	if (w)
+	{
+	    scaleaddonRenderWindowTitle (w);
+	    addWindowDamage (w);
+	}
 
-	ad->lastHoveredWindow = sd->hoveredWindow;
+	ad->lastHighlightedWindow = ad->highlightedWindow;
     }
 }
 
-static CompWindow *
-scaleaddonCheckForWindowAt (CompScreen * s, int x, int y)
+static CompWindow*
+scaleaddonCheckForWindowAt (CompScreen *s,
+			    int        x,
+			    int        y)
 {
     float      x1, y1, x2, y2;
     CompWindow *w;
 
     for (w = s->reverseWindows; w; w = w->prev)
     {
-        SCALE_WINDOW(w);
+        SCALE_WINDOW (w);
 
         if (sw->slot)
 	{
@@ -449,8 +493,8 @@ scaleaddonZoomWindow (CompDisplay     *d,
 	    ADDON_WINDOW (w);
 
 	    XRectangle outputRect;
-	    BOX outputBox;
-	    int head;
+	    BOX        outputBox;
+	    int        head;
 
 	    if (!sw->slot)
 		return FALSE;
@@ -466,7 +510,7 @@ scaleaddonZoomWindow (CompDisplay     *d,
 	    if (!aw->rescaled)
 	    {
 		aw->oldAbove = w->next;
-		raiseWindow(w);
+		raiseWindow (w);
 	
 		/* backup old values */
 		aw->origSlot = *sw->slot;
@@ -519,9 +563,7 @@ scaleaddonHandleEvent (CompDisplay *d,
     {
     case PropertyNotify:
 	{
-	    SCALE_DISPLAY (d);
-	    if (event->xproperty.window == sd->hoveredWindow &&
-		event->xproperty.atom == XA_WM_NAME)
+	    if (event->xproperty.atom == XA_WM_NAME)
 	    {
 		CompWindow *w;
 
@@ -547,7 +589,12 @@ scaleaddonHandleEvent (CompDisplay *d,
 	    {
 		SCALE_SCREEN (s);
 		if (ss->grabIndex)
-		    scaleaddonCheckHoveredWindow (s);
+		{
+		    SCALE_DISPLAY (d);
+
+		    ad->highlightedWindow = sd->hoveredWindow;
+		    scaleaddonCheckWindowHighlight (s);
+		}
 	    }
 	}
 	break;
@@ -563,35 +610,86 @@ scaleaddonScalePaintDecoration (CompWindow              *w,
 				Region                  region,
 				unsigned int            mask)
 {
-    ADDON_SCREEN (w->screen);
-    SCALE_SCREEN (w->screen);
-    SCALE_DISPLAY (w->screen->display);
+    CompScreen *s = w->screen;
+
+    ADDON_SCREEN (s);
+    SCALE_SCREEN (s);
+    ADDON_WINDOW (w);
 
     UNWRAP (as, ss, scalePaintDecoration);
     (*ss->scalePaintDecoration) (w, attrib, transform, region, mask);
     WRAP (as, ss, scalePaintDecoration, scaleaddonScalePaintDecoration);
 
-    scaleaddonCheckHoveredWindow (w->screen);
-
-    if ((w->id == sd->hoveredWindow) &&
-	((ss->state == SCALE_STATE_WAIT) || (ss->state == SCALE_STATE_OUT)))
+    if ((ss->state == SCALE_STATE_WAIT) || (ss->state == SCALE_STATE_OUT))
     {
-	if (scaleaddonGetWindowHighlight (w->screen))
-	    scaleaddonDrawWindowHighlight (w);
+	if (scaleaddonGetWindowHighlight (s))
+	{
+	    ADDON_DISPLAY (s->display);
 
-	if (as->textPixmap)
+	    if (w->id == ad->highlightedWindow)
+		scaleaddonDrawWindowHighlight (w);
+	}
+
+	if (aw->textPixmap)
 	    scaleaddonDrawWindowTitle (w);
     }
 }
 
+static void
+scaleaddonSelectWindow (CompWindow *w)
+{
+    CompScreen *s = w->screen;
+
+    ADDON_DISPLAY (s->display);
+    ADDON_SCREEN (s);
+    SCALE_SCREEN (s);
+
+    ad->highlightedWindow = w->id;
+    scaleaddonCheckWindowHighlight (s);
+
+    UNWRAP (as, ss, selectWindow);
+    (*ss->selectWindow) (w);
+    WRAP (as, ss, selectWindow, scaleaddonSelectWindow);
+}
+
+static void
+scaleaddonDonePaintScreen (CompScreen *s)
+{
+    ADDON_SCREEN (s);
+    SCALE_SCREEN (s);
+
+    if (ss->state != SCALE_STATE_NONE &&
+	as->lastState == SCALE_STATE_NONE)
+    {
+	CompWindow *w;
+
+	for (w = s->windows; w; w = w->next)
+	    scaleaddonRenderWindowTitle (w);
+    }
+    else if (ss->state == SCALE_STATE_NONE &&
+	     as->lastState != SCALE_STATE_NONE)
+    {
+	CompWindow *w;
+
+	for (w = s->windows; w; w = w->next)
+	    scaleaddonFreeWindowTitle (w);
+    }
+
+    as->lastState = ss->state;
+
+    UNWRAP (as, s, donePaintScreen);
+    (*s->donePaintScreen) (s);
+    WRAP (as, s, donePaintScreen, scaleaddonDonePaintScreen);
+}
+
 static void
 scaleaddonHandleCompizEvent (CompDisplay *d,
-			     char        *pluginName,
-			     char        *eventName,
+			     const char  *pluginName,
+			     const char  *eventName,
 			     CompOption  *option,
 			     int         nOption)
 {
-    ADDON_DISPLAY(d);
+    ADDON_DISPLAY (d);
 
     UNWRAP (ad, d, handleCompizEvent);
     (*d->handleCompizEvent) (d, pluginName, eventName, option, nOption);
@@ -600,16 +698,30 @@ scaleaddonHandleCompizEvent (CompDisplay *d,
     if ((strcmp (pluginName, "scale") == 0) &&
 	(strcmp (eventName, "activate") == 0))
     {
-	Window xid = getIntOptionNamed (option, nOption, "root", 0);
-	Bool activated = getIntOptionNamed (option, nOption, "activated", FALSE);
-	CompScreen *s = findScreenAtDisplay (d, xid);
+	Window     xid;
+	Bool       activated;
+	CompScreen *s;
+	
+	xid = getIntOptionNamed (option, nOption, "root", 0);
+	activated = getBoolOptionNamed (option, nOption, "active", FALSE);
+	s = findScreenAtDisplay (d, xid);
 
 	if (s)
 	{
 	    if (activated)
 	    {
-		addScreenAction (s, scaleaddonGetClose (s->display));
-		addScreenAction (s, scaleaddonGetZoom (s->display));
+		addScreenAction (s, scaleaddonGetCloseKey (d));
+		addScreenAction (s, scaleaddonGetZoomKey (d));
+		addScreenAction (s, scaleaddonGetCloseButton (d));
+		addScreenAction (s, scaleaddonGetZoomButton (d));
+
+		/* TODO: or better
+		   ad->highlightedWindow     = sd->selectedWindow;
+		   here? do we want to show up the highlight without
+		   mouse move initially? */
+		ad->highlightedWindow     = None;
+		ad->lastHighlightedWindow = None;
+		scaleaddonCheckWindowHighlight (s);
 	    }
 	    else
 	    {
@@ -621,8 +733,10 @@ scaleaddonHandleCompizEvent (CompDisplay *d,
 		    aw->rescaled = FALSE;
 		}
 
-		removeScreenAction (s, scaleaddonGetClose (s->display));
-		removeScreenAction (s, scaleaddonGetZoom (s->display));
+		removeScreenAction (s, scaleaddonGetCloseKey (d));
+		removeScreenAction (s, scaleaddonGetZoomKey (d));
+		removeScreenAction (s, scaleaddonGetCloseButton (d));
+		removeScreenAction (s, scaleaddonGetZoomButton (d));
 	    }
 	}
     }
@@ -635,196 +749,217 @@ scaleaddonHandleCompizEvent (CompDisplay *d,
 #define ORGANIC_STEP 0.05
 
 static int
-organicCompareWindows (const void *elem1, const void *elem2)
+organicCompareWindows (const void *elem1,
+		       const void *elem2)
 {
     CompWindow *w1 = *((CompWindow **) elem1);
     CompWindow *w2 = *((CompWindow **) elem2);
 
-    return (WIN_X(w1) + WIN_Y(w1)) - (WIN_X(w2) + WIN_Y(w2));
+    return (WIN_X (w1) + WIN_Y (w1)) - (WIN_X (w2) + WIN_Y (w2));
 }
 
 static double
-layoutOrganicCalculateOverlap (CompScreen * s, int win, int x, int y)
+layoutOrganicCalculateOverlap (CompScreen *s,
+			       int        win,
+			       int        x,
+			       int        y)
 {
+    int    i;
+    int    x1, y1, x2, y2;
+    int    overlapX, overlapY;
+    int    xMin, xMax, yMin, yMax;
+    double result = -0.01;
+
     SCALE_SCREEN (s);
     ADDON_SCREEN (s);
 
-    int i;
-    int x1 = x;
-    int y1 = y;
-    int x2 = x1 + WIN_W (ss->windows[win]) * as->scale;
-    int y2 = y1 + WIN_H (ss->windows[win]) * as->scale;
-    int overlapx = 0, overlapy = 0;
-    int xmin, xmax;
-    int ymin, ymax;
-
-    double result = -0.01;
+    x1 = x;
+    y1 = y;
+    x2 = x1 + WIN_W (ss->windows[win]) * as->scale;
+    y2 = y1 + WIN_H (ss->windows[win]) * as->scale;
 
     for (i = 0; i < ss->nWindows; i++)
     {
 	if (i == win)
 	    continue;
-	overlapx = overlapy = 0;
-	xmax = MAX (ss->slots[i].x1, x1);
-	xmin = MIN (ss->slots[i].x1 + WIN_W(ss->windows[i]) * as->scale, x2);
-	if (xmax <= xmin)
-	    overlapx = xmin - xmax;
 
-	ymax = MAX (ss->slots[i].y1, y1);
-	ymin = MIN (ss->slots[i].y1 + WIN_H(ss->windows[i]) * as->scale, y2);
+	overlapX = overlapY = 0;
+	xMax = MAX (ss->slots[i].x1, x1);
+	xMin = MIN (ss->slots[i].x1 + WIN_W (ss->windows[i]) * as->scale, x2);
+	if (xMax <= xMin)
+	    overlapX = xMin - xMax;
 
-	if (ymax <= ymin)
-	    overlapy = ymin - ymax;
+	yMax = MAX (ss->slots[i].y1, y1);
+	yMin = MIN (ss->slots[i].y1 + WIN_H (ss->windows[i]) * as->scale, y2);
 
-	result += (double)overlapx * overlapy;
+	if (yMax <= yMin)
+	    overlapY = yMin - yMax;
+
+	result += (double)overlapX * overlapY;
     }
+
     return result;
 }
 
 static double
-layoutOrganicFindBestHorizontalPosition (CompScreen * s, int win,
-					 int *bestx, int areaWidth)
+layoutOrganicFindBestHorizontalPosition (CompScreen *s,
+					 int        win,
+					 int        *bestX,
+					 int        areaWidth)
 {
+    int    i, y1, y2, w;
+    double bestOverlap = 1e31, overlap;
+
     SCALE_SCREEN (s);
     ADDON_SCREEN (s);
 
-    int i;
-    int y1 = ss->slots[win].y1;
-    int y2 = ss->slots[win].y1 + WIN_H(ss->windows[win]) * as->scale;
+    y1 = ss->slots[win].y1;
+    y2 = ss->slots[win].y1 + WIN_H (ss->windows[win]) * as->scale;
 
-    double bestoverlap = 1e31, overlap;
-    int w = WIN_W(ss->windows[win]) * as->scale;
-
-    *bestx = ss->slots[win].x1;
+    w = WIN_W (ss->windows[win]) * as->scale;
+    *bestX = ss->slots[win].x1;
 
     for (i = 0; i < ss->nWindows; i++)
     {
+	CompWindow *lw = ss->windows[i];
 	if (i == win)
 	    continue;
 
 	if (ss->slots[i].y1 < y2 &&
-	    ss->slots[i].y1 + WIN_H(ss->windows[i]) * as->scale > y1)
+	    ss->slots[i].y1 + WIN_H (lw) * as->scale > y1)
 	{
 	    if (ss->slots[i].x1 - w >= 0)
 	    {
-		double overlap = layoutOrganicCalculateOverlap (s, win,
-								ss->slots[i].x1 - w, y1);
+		double overlap;
+		
+		overlap = layoutOrganicCalculateOverlap (s, win,
+		 					 ss->slots[i].x1 - w,
+							 y1);
 
-		if (overlap < bestoverlap)
+		if (overlap < bestOverlap)
 		{
-		    *bestx = ss->slots[i].x1 - w;
-		    bestoverlap = overlap;
+		    *bestX = ss->slots[i].x1 - w;
+		    bestOverlap = overlap;
 		}
 	    }
-	    if (WIN_W(ss->windows[i]) * as->scale +
-		ss->slots[i].x1 + w < areaWidth)
+	    if (WIN_W (lw) * as->scale + ss->slots[i].x1 + w < areaWidth)
 	    {
-		double overlap = layoutOrganicCalculateOverlap (s, win,
-								ss->slots[i].x1 +
-								WIN_W(ss->windows[i]) *
-		     						as->scale, y1);
+		double overlap;
+		
+		overlap = layoutOrganicCalculateOverlap (s, win,
+		 					 ss->slots[i].x1 +
+		 					 WIN_W (lw) * as->scale,
+		 					 y1);
 
-		if (overlap < bestoverlap)
+		if (overlap < bestOverlap)
 		{
-		    *bestx = ss->slots[i].x1 + WIN_W(ss->windows[i]) * as->scale;
-		    bestoverlap = overlap;
+		    *bestX = ss->slots[i].x1 + WIN_W (lw) * as->scale;
+		    bestOverlap = overlap;
 		}
 	    }
 	}
     }
 
     overlap = layoutOrganicCalculateOverlap (s, win, 0, y1);
-    if (overlap < bestoverlap)
+    if (overlap < bestOverlap)
     {
-	*bestx = 0;
-	bestoverlap = overlap;
+	*bestX = 0;
+	bestOverlap = overlap;
     }
 
     overlap = layoutOrganicCalculateOverlap (s, win, areaWidth - w, y1);
-    if (overlap < bestoverlap)
+    if (overlap < bestOverlap)
     {
-	*bestx = areaWidth - w;
-	bestoverlap = overlap;
+	*bestX = areaWidth - w;
+	bestOverlap = overlap;
     }
 
-    return bestoverlap;
+    return bestOverlap;
 }
 
 static double
-layoutOrganicFindBestVerticalPosition (CompScreen * s, int win,
-				       int *besty, int areaHeight)
+layoutOrganicFindBestVerticalPosition (CompScreen *s,
+				       int        win,
+				       int        *bestY,
+				       int        areaHeight)
 {
+    int    i, x1, x2, h;
+    double bestOverlap = 1e31, overlap;
+
     SCALE_SCREEN (s);
     ADDON_SCREEN (s);
 
-    int i;
-    int x1 = ss->slots[win].x1;
-    int x2 = ss->slots[win].x1 + WIN_W(ss->windows[win]) * as->scale;
-
-    double bestoverlap = 1e31, overlap;
-    int h = WIN_H(ss->windows[win]) * as->scale;
-
-    *besty = ss->slots[win].y1;
+    x1 = ss->slots[win].x1;
+    x2 = ss->slots[win].x1 + WIN_W (ss->windows[win]) * as->scale;
+    h = WIN_H (ss->windows[win]) * as->scale;
+    *bestY = ss->slots[win].y1;
 
     for (i = 0; i < ss->nWindows; i++)
     {
+	CompWindow *w = ss->windows[i];
+
 	if (i == win)
 	    continue;
 
 	if (ss->slots[i].x1 < x2 &&
-	    ss->slots[i].x1 + WIN_W(ss->windows[i]) * as->scale > x1)
+	    ss->slots[i].x1 + WIN_W (w) * as->scale > x1)
 	{
 	    if (ss->slots[i].y1 - h >= 0 && ss->slots[i].y1 < areaHeight)
 	    {
-		double overlap = layoutOrganicCalculateOverlap (s, win, x1,
-								ss->slots[i].y1 - h);
-		if (overlap < bestoverlap)
+		double overlap;
+		overlap = layoutOrganicCalculateOverlap (s, win, x1,
+	 						 ss->slots[i].y1 - h);
+		if (overlap < bestOverlap)
 		{
-		    *besty = ss->slots[i].y1 - h;
-		    bestoverlap = overlap;
+		    *bestY = ss->slots[i].y1 - h;
+		    bestOverlap = overlap;
 		}
 	    }
-	    if (WIN_H(ss->windows[i]) * as->scale + ss->slots[i].y1 > 0 &&
-		WIN_H(ss->windows[i]) * as->scale + h + ss->slots[i].y1 < areaHeight)
+	    if (WIN_H (w) * as->scale + ss->slots[i].y1 > 0 &&
+		WIN_H (w) * as->scale + h + ss->slots[i].y1 < areaHeight)
 	    {
-		double overlap = layoutOrganicCalculateOverlap (s, win, x1,
-								WIN_H(ss->windows[i]) *
-		     						as->scale +
-								ss->slots[i].y1);
+		double overlap;
+		
+		overlap = layoutOrganicCalculateOverlap (s, win, x1,
+		 					 WIN_H (w) * as->scale +
+							 ss->slots[i].y1);
 
-		if (overlap < bestoverlap)
+		if (overlap < bestOverlap)
 		{
-		    *besty = ss->slots[i].y1 + WIN_H(ss->windows[i]) * as->scale;
-		    bestoverlap = overlap;
+		    *bestY = ss->slots[i].y1 + WIN_H(w) * as->scale;
+		    bestOverlap = overlap;
 		}
 	    }
 	}
     }
 
     overlap = layoutOrganicCalculateOverlap (s, win, x1, 0);
-    if (overlap < bestoverlap)
+    if (overlap < bestOverlap)
     {
-	*besty = 0;
-	bestoverlap = overlap;
+	*bestY = 0;
+	bestOverlap = overlap;
     }
 
     overlap = layoutOrganicCalculateOverlap (s, win, x1, areaHeight - h);
-    if (overlap < bestoverlap)
+    if (overlap < bestOverlap)
     {
-	*besty = areaHeight - h;
-	bestoverlap = overlap;
+	*bestY = areaHeight - h;
+	bestOverlap = overlap;
     }
 
-    return bestoverlap;
+    return bestOverlap;
 }
 
 static Bool
-layoutOrganicLocalSearch (CompScreen * s, int areaWidth, int areaHeight)
+layoutOrganicLocalSearch (CompScreen *s,
+			  int        areaWidth,
+			  int        areaHeight)
 {
-    SCALE_SCREEN(s);
-    Bool improvement;
-    int i;
-    double totaloverlap;
+    Bool   improvement;
+    int    i;
+    double totalOverlap;
+
+    SCALE_SCREEN (s);
 
     do
     {
@@ -835,51 +970,54 @@ layoutOrganicLocalSearch (CompScreen * s, int areaWidth, int areaHeight)
 
 	    do
 	    {
-		int newx, newy;
-		double oldoverlap, overlaph, overlapv;
+		int    newX, newY;
+		double oldOverlap, overlapH, overlapV;
 
 		improved = FALSE;
-		oldoverlap = layoutOrganicCalculateOverlap (s, i,
+		oldOverlap = layoutOrganicCalculateOverlap (s, i,
  							    ss->slots[i].x1,
 							    ss->slots[i].y1);
 
-		overlaph =
-		    layoutOrganicFindBestHorizontalPosition (s, i, &newx, areaWidth);
-		overlapv =
-		    layoutOrganicFindBestVerticalPosition (s, i, &newy, areaHeight);
+		overlapH = layoutOrganicFindBestHorizontalPosition (s, i,
+								    &newX,
+								    areaWidth);
+		overlapV = layoutOrganicFindBestVerticalPosition (s, i,
+								  &newY,
+								  areaHeight);
 
-		if (overlaph < oldoverlap - 0.1 ||
-		    overlapv < oldoverlap - 0.1)
+		if (overlapH < oldOverlap - 0.1 ||
+		    overlapV < oldOverlap - 0.1)
 		{
 		    improved = TRUE;
 		    improvement = TRUE;
-		    if (overlapv > overlaph)
-			ss->slots[i].x1 = newx;
+		    if (overlapV > overlapH)
+			ss->slots[i].x1 = newX;
 		    else
-			ss->slots[i].y1 = newy;
+			ss->slots[i].y1 = newY;
 		}
-
     	    }
 	    while (improved);
 	}
     }
     while (improvement);
 
-    totaloverlap = 0.0;
+    totalOverlap = 0.0;
     for (i = 0; i < ss->nWindows; i++)
     {
-	totaloverlap += layoutOrganicCalculateOverlap (s,
-						       i, ss->slots[i].x1,
+	totalOverlap += layoutOrganicCalculateOverlap (s, i,
+						       ss->slots[i].x1,
 						       ss->slots[i].y1);
     }
-    return (totaloverlap > 0.1);
+
+    return (totalOverlap > 0.1);
 }
 
 static void
-layoutOrganicRemoveOverlap (CompScreen * s, int areaWidth, int areaHeight)
+layoutOrganicRemoveOverlap (CompScreen *s,
+			    int        areaWidth,
+			    int        areaHeight)
 {
-    int i;
-    int spacing;
+    int        i, spacing;
     CompWindow *w;
 
     SCALE_SCREEN (s);
@@ -896,32 +1034,30 @@ layoutOrganicRemoveOverlap (CompScreen * s, int areaWidth, int areaHeight)
 
 	    w = ss->windows[i];
 
-	    centerX = ss->slots[i].x1 + WIN_W(w) / 2;
-	    centerY = ss->slots[i].y1 + WIN_H(w) / 2;
+	    centerX = ss->slots[i].x1 + WIN_W (w) / 2;
+	    centerY = ss->slots[i].y1 + WIN_H (w) / 2;
 
 	    newWidth = (int)((1.0 - ORGANIC_STEP) *
-			     (double)WIN_W(w)) - spacing / 2;
+			     (double)WIN_W (w)) - spacing / 2;
 	    newHeight = (int)((1.0 - ORGANIC_STEP) *
-			      (double)WIN_H(w)) - spacing / 2;
+			      (double)WIN_H (w)) - spacing / 2;
 	    newX = centerX - (newWidth / 2);
 	    newY = centerY - (newHeight / 2);
 
 	    ss->slots[i].x1 = newX;
 	    ss->slots[i].y1 = newY;
-	    ss->slots[i].x2 = newX + WIN_W(w);
-	    ss->slots[i].y2 = newY + WIN_H(w);
+	    ss->slots[i].x2 = newX + WIN_W (w);
+	    ss->slots[i].y2 = newY + WIN_H (w);
 	}
 	as->scale -= ORGANIC_STEP;
     }
 }
 
 static Bool
-layoutOrganicThumbs(CompScreen * s)
+layoutOrganicThumbs (CompScreen *s)
 {
     CompWindow *w;
-
-    int i;
-    int moMode;
+    int        i, moMode;
     XRectangle workArea;
 
     SCALE_SCREEN (s);
@@ -946,8 +1082,8 @@ layoutOrganicThumbs(CompScreen * s)
 
     for (i = 0; i < ss->nWindows; i++)
     {
-	SCALE_WINDOW (ss->windows[i]);
 	w = ss->windows[i];
+	SCALE_WINDOW (w);
 
 	sw->slot = &ss->slots[i];
 	ss->slots[i].x1 = WIN_X (w) - workArea.x;
@@ -985,15 +1121,16 @@ layoutOrganicThumbs(CompScreen * s)
 				workArea.height - workArea.y);
     for (i = 0; i < ss->nWindows; i++)
     {
-	SCALE_WINDOW (ss->windows[i]);
+	w = ss->windows[i];
+	SCALE_WINDOW (w);
 
 	if (ss->type == ScaleTypeGroup)
 	    raiseWindow (ss->windows[i]);
 
-	ss->slots[i].x1 += ss->windows[i]->input.left + workArea.x;
-	ss->slots[i].x2 += ss->windows[i]->input.left + workArea.x;
-	ss->slots[i].y1 += ss->windows[i]->input.top + workArea.y;
-	ss->slots[i].y2 += ss->windows[i]->input.top + workArea.y;
+	ss->slots[i].x1 += w->input.left + workArea.x;
+	ss->slots[i].x2 += w->input.left + workArea.x;
+	ss->slots[i].y1 += w->input.top + workArea.y;
+	ss->slots[i].y2 += w->input.top + workArea.y;
 	sw->adjust = TRUE;
     }
 
@@ -1032,16 +1169,22 @@ scaleaddonScreenOptionChanged (CompScreen              *s,
 {
     switch (num)
     {
+	case ScaleaddonScreenOptionWindowTitle:
 	case ScaleaddonScreenOptionTitleBold:
 	case ScaleaddonScreenOptionTitleSize:
 	case ScaleaddonScreenOptionBorderSize:
 	case ScaleaddonScreenOptionFontColor:
 	case ScaleaddonScreenOptionBackColor:
 	    {
-		ADDON_DISPLAY (s->display);
+		CompWindow *w;
 
-		scaleaddonFreeWindowTitle (s);
-		ad->lastHoveredWindow = None;
+		for (w = s->windows; w; w = w->next)
+		{
+		    ADDON_WINDOW (w);
+
+		    if (aw->textPixmap)
+			scaleaddonRenderWindowTitle (w);
+		}
 	    }
 	    break;
 	default:
@@ -1054,24 +1197,14 @@ scaleaddonInitDisplay (CompPlugin  *p,
 	      	       CompDisplay *d)
 {
     ScaleAddonDisplay *ad;
-    CompPlugin        *scale = findActivePlugin ("scale");
-    CompOption        *option;
-    int               nOption;
 
-    if (!scale || !scale->vTable->getDisplayOptions)
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
 	return FALSE;
 
-    option = (*scale->vTable->getDisplayOptions) (scale, d, &nOption);
-
-    if (getIntOptionNamed (option, nOption, "abi", 0) != SCALE_ABIVERSION)
-    {
-	compLogMessage (d, "scaleaddon", CompLogLevelError,
-			"scale ABI version mismatch");
+    if (!checkPluginABI ("scale", SCALE_ABIVERSION))
 	return FALSE;
-    }
 
-    scaleDisplayPrivateIndex = getIntOptionNamed (option, nOption, "index", -1);
-    if (scaleDisplayPrivateIndex < 0)
+    if (!getPluginDisplayIndex (d, "scale", &scaleDisplayPrivateIndex))
 	return FALSE;
 
     ad = malloc (sizeof (ScaleAddonDisplay));
@@ -1085,15 +1218,23 @@ scaleaddonInitDisplay (CompPlugin  *p,
 	return FALSE;
     }
 
+    ad->textAvailable = checkPluginABI ("text", TEXT_ABIVERSION);
+    if (!ad->textAvailable)
+	compLogMessage (d, "scaleaddon", CompLogLevelWarn,
+			"No compatible text plugin found.");
+
     WRAP (ad, d, handleEvent, scaleaddonHandleEvent);
     WRAP (ad, d, handleCompizEvent, scaleaddonHandleCompizEvent);
 
-    d->privates[displayPrivateIndex].ptr = ad;
+    d->base.privates[displayPrivateIndex].ptr = ad;
 
-    ad->lastHoveredWindow = None;
+    ad->highlightedWindow     = None;
+    ad->lastHighlightedWindow = None;
 
-    scaleaddonSetCloseInitiate (d, scaleaddonCloseWindow);
-    scaleaddonSetZoomInitiate (d, scaleaddonZoomWindow);
+    scaleaddonSetCloseKeyInitiate (d, scaleaddonCloseWindow);
+    scaleaddonSetZoomKeyInitiate (d, scaleaddonZoomWindow);
+    scaleaddonSetCloseButtonInitiate (d, scaleaddonCloseWindow);
+    scaleaddonSetZoomButtonInitiate (d, scaleaddonZoomWindow);
 
     return TRUE;
 }
@@ -1114,7 +1255,7 @@ scaleaddonFiniDisplay (CompPlugin  *p,
 
 static Bool
 scaleaddonInitScreen (CompPlugin *p,
-	      	      CompScreen *s)
+		      CompScreen *s)
 {
     ScaleAddonScreen *as;
 
@@ -1134,20 +1275,20 @@ scaleaddonInitScreen (CompPlugin *p,
 
     as->scale = 1.0f;
 
-    as->textPixmap = None;
-    initTexture (s, &as->textTexture);
-
+    WRAP (as, s, donePaintScreen, scaleaddonDonePaintScreen);
     WRAP (as, ss, scalePaintDecoration, scaleaddonScalePaintDecoration);
+    WRAP (as, ss, selectWindow, scaleaddonSelectWindow);
     WRAP (as, ss, layoutSlotsAndAssignWindows,
 	  scaleaddonLayoutSlotsAndAssignWindows);
 
+    scaleaddonSetWindowTitleNotify (s, scaleaddonScreenOptionChanged);
     scaleaddonSetTitleBoldNotify (s, scaleaddonScreenOptionChanged);
     scaleaddonSetTitleSizeNotify (s, scaleaddonScreenOptionChanged);
     scaleaddonSetBorderSizeNotify (s, scaleaddonScreenOptionChanged);
     scaleaddonSetFontColorNotify (s, scaleaddonScreenOptionChanged);
     scaleaddonSetBackColorNotify (s, scaleaddonScreenOptionChanged);
 
-    s->privates[ad->screenPrivateIndex].ptr = as;
+    s->base.privates[ad->screenPrivateIndex].ptr = as;
 
     return TRUE;
 }
@@ -1159,11 +1300,11 @@ scaleaddonFiniScreen (CompPlugin *p,
     ADDON_SCREEN (s);
     SCALE_SCREEN (s);
 
+    UNWRAP (as, s, donePaintScreen);
     UNWRAP (as, ss, scalePaintDecoration);
+    UNWRAP (as, ss, selectWindow);
     UNWRAP (as, ss, layoutSlotsAndAssignWindows);
 
-    scaleaddonFreeWindowTitle (s);
-
     freeWindowPrivateIndex (s, as->windowPrivateIndex);
     free (as);
 }
@@ -1182,7 +1323,12 @@ scaleaddonInitWindow (CompPlugin *p,
 
     aw->rescaled = FALSE;
 
-    w->privates[as->windowPrivateIndex].ptr = aw;
+    w->base.privates[as->windowPrivateIndex].ptr = aw;
+
+    aw->textPixmap = None;
+    initTexture (w->screen, &aw->textTexture);
+    aw->textWidth  = 0;
+    aw->textHeight = 0;
 
     return TRUE;
 }
@@ -1193,9 +1339,39 @@ scaleaddonFiniWindow (CompPlugin *p,
 {
     ADDON_WINDOW (w);
 
+    scaleaddonFreeWindowTitle (w);
+
     free (aw);
 }
 
+static CompBool
+scaleaddonInitObject (CompPlugin *p,
+		      CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) scaleaddonInitDisplay,
+	(InitPluginObjectProc) scaleaddonInitScreen,
+	(InitPluginObjectProc) scaleaddonInitWindow
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+scaleaddonFiniObject (CompPlugin *p,
+		      CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) scaleaddonFiniDisplay,
+	(FiniPluginObjectProc) scaleaddonFiniScreen,
+	(FiniPluginObjectProc) scaleaddonFiniWindow
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
 static Bool
 scaleaddonInit (CompPlugin *p)
 {
@@ -1212,27 +1388,13 @@ scaleaddonFini (CompPlugin *p)
     freeDisplayPrivateIndex (displayPrivateIndex);
 }
 
-static int
-scaleaddonGetVersion (CompPlugin *plugin,
-	      	      int	 version)
-{
-    return ABIVERSION;
-}
-
 CompPluginVTable scaleaddonVTable = {
     "scaleaddon",
-    scaleaddonGetVersion,
     0,
     scaleaddonInit,
     scaleaddonFini,
-    scaleaddonInitDisplay,
-    scaleaddonFiniDisplay,
-    scaleaddonInitScreen,
-    scaleaddonFiniScreen,
-    scaleaddonInitWindow,
-    scaleaddonFiniWindow,
-    0,
-    0,
+    scaleaddonInitObject,
+    scaleaddonFiniObject,
     0,
     0
 };
