@@ -1,6 +1,13 @@
 /*
- * text.c - adds text image support to beryl.
- * Copyright: (C) 2006 Patrick Niklaus
+ * Compiz text plugin
+ * Description: Adds text to pixmap support to Compiz.
+ *
+ * text.c
+ *
+ * Copyright: (C) 2006-2007 Patrick Niklaus, Danny Baumann, Dennis Kasprzyk
+ * Authors: Patrick Niklaus <marex@opencompsiting.org>
+ *	    Danny Baumann   <maniac@opencompositing.org>
+ *	    Dennis Kasprzyk <onestone@opencompositing.org>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -30,8 +37,9 @@
 #include <pango/pango.h>
 #include <pango/pangocairo.h>
 
-#include <compiz.h>
-#include "text.h"
+#include <compiz-core.h>
+#include "compiz-text.h"
+#include "text_options.h"
 
 static int displayPrivateIndex;
 
@@ -43,7 +51,7 @@ typedef struct _TextDisplay
 } TextDisplay;
 
 #define GET_TEXT_DISPLAY(d)				    \
-    ((TextDisplay *) (d)->privates[displayPrivateIndex].ptr)
+    ((TextDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define TEXT_DISPLAY(d)			 \
     TextDisplay *td = GET_TEXT_DISPLAY (d)
@@ -159,6 +167,7 @@ textFileToImage (CompDisplay *d,
 	Display              *dpy = d->display;
 	Screen               *screen;
 	int                  w, h;
+	char                 *text = NULL;
 	
 	textAttrib = (CompTextAttrib*) name; /* get it through the name */
 	screen = ScreenOfDisplay (dpy, textAttrib->screen->screenNum);
@@ -254,49 +263,73 @@ textFileToImage (CompDisplay *d,
 
 	switch (textAttrib->renderMode) {
 	case TextRenderNormal:
-	    {
-		char *text = (char*) textAttrib->data;
-		pango_layout_set_text (layout, text, -1);
-	    }
+	    text = strdup ((char*) textAttrib->data);
 	    break;
 	case TextRenderWindowTitle:
 	    {
 		Window xid = (Window) textAttrib->data;
-		char   *text = textGetWindowName (d, xid);
+		text = textGetWindowName (d, xid);
+	    }
+	    break;
+	case TextRenderWindowTitleWithViewport:
+	    {
+		Window xid = (Window) textAttrib->data;
+		char *title = textGetWindowName (d, xid);
 
-		if (text)
+		if (title)
 		{
-		    pango_layout_set_text (layout, text, -1);
-		    free (text);
+		    CompWindow *w;
+
+		    w = findWindowAtDisplay (d, xid);
+		    if (w)
+		    {
+			int vx, vy, viewport;
+
+			defaultViewportForWindow (w, &vx, &vy);
+			viewport = vy * w->screen->hsize + vx + 1;
+			asprintf (&text, "%s -[%d]-", title, viewport);
+			free (title);
+		    }
+		    else
+			text = title;
 		}
-		else
-		{
-		    pango_font_description_free (font);
-		    g_object_unref (layout);
-		    cairo_destroy (cr);
-		    cairo_surface_destroy (surface);
-		    XFreePixmap (dpy, pixmap);
-		    return FALSE;
-		}
-    	    }
+	    }
 	    break;
 	default:
 	    break;
 	}
 
+	if (text)
+	{
+	    pango_layout_set_text (layout, text, -1);
+	    free (text);
+	}
+	else
+	{
+	    pango_font_description_free (font);
+	    g_object_unref (layout);
+	    cairo_destroy (cr);
+	    cairo_surface_destroy (surface);
+	    XFreePixmap (dpy, pixmap);
+	    return FALSE;
+	}
+
 	pango_layout_get_pixel_size (layout, &w, &h);
 
-	w = MIN (textAttrib->maxwidth, w);
-	h = MIN (textAttrib->maxheight, h);
+	w = MIN (textAttrib->maxWidth, w);
+	h = MIN (textAttrib->maxHeight, h);
 
 	/* update the size of the pango layout */
-	pango_layout_set_width (layout, textAttrib->maxwidth * PANGO_SCALE);
+	pango_layout_set_width (layout, textAttrib->maxWidth * PANGO_SCALE);
 
 	cairo_surface_destroy (surface);
 	cairo_destroy (cr);
 	XFreePixmap (dpy, pixmap);
+	pixmap = None;
+
+	if (w > 0 && h > 0)
+	    pixmap = XCreatePixmap (dpy, textAttrib->screen->root, w, h, 32);
 
-	pixmap = XCreatePixmap (dpy, textAttrib->screen->root, w, h, 32);
 	if (!pixmap)
 	{
 	    compLogMessage (d, "text", CompLogLevelError,
@@ -371,6 +404,10 @@ textInitDisplay (CompPlugin  *p,
 		 CompDisplay *d)
 {
     TextDisplay *td;
+    CompOption  *o;
+
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
 
     td = malloc (sizeof (TextDisplay));
     if (!td)
@@ -381,7 +418,10 @@ textInitDisplay (CompPlugin  *p,
 
     WRAP (td, d, fileToImage, textFileToImage);
 
-    d->privates[displayPrivateIndex].ptr = td;
+    d->base.privates[displayPrivateIndex].ptr = td;
+
+    o = textGetAbiOption (d);
+    o->value.i = TEXT_ABIVERSION;
 
     return TRUE;
 }
@@ -397,6 +437,29 @@ textFiniDisplay (CompPlugin  *p,
     free (td);
 }
 
+static CompBool
+textInitObject (CompPlugin *p,
+		CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) textInitDisplay
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+textFiniObject (CompPlugin *p,
+		CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) textFiniDisplay
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
 
 static Bool
 textInit (CompPlugin *p)
@@ -414,27 +477,13 @@ textFini (CompPlugin *p)
     freeDisplayPrivateIndex(displayPrivateIndex);
 }
 
-static int
-textGetVersion (CompPlugin *p,
-		int        version)
-{
-    return ABIVERSION;
-}
-
 CompPluginVTable textVTable = {
     "text",
-    textGetVersion,
     0,
     textInit,
     textFini,
-    textInitDisplay,
-    textFiniDisplay,
-    0,
-    0,
-    0,
-    0,
-    0,
-    0,
+    textInitObject,
+    textFiniObject,
     0,
     0
 };
@@ -444,3 +493,4 @@ getCompPluginInfo (void)
 {
     return &textVTable;
 }
+
