@@ -13,11 +13,13 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ * along with this program; if not, write to the
+ * Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor,
+ * Boston, MA  02110-1301, USA.
  */
 
-#include <compiz.h>
+#include <compiz-core.h>
 
 #include <X11/Xatom.h>
 
@@ -34,7 +36,7 @@
 #define WINRULES_SCREEN_OPTION_BELOW_MATCH        3
 #define WINRULES_SCREEN_OPTION_STICKY_MATCH       4
 #define WINRULES_SCREEN_OPTION_FULLSCREEN_MATCH   5
-#define WINRULES_SCREEN_OPTION_WIDGET_MATCH       6
+#define WINRULES_SCREEN_OPTION_NOARGB_MATCH       6
 #define WINRULES_SCREEN_OPTION_NOMOVE_MATCH       7
 #define WINRULES_SCREEN_OPTION_NORESIZE_MATCH     8
 #define WINRULES_SCREEN_OPTION_NOMINIMIZE_MATCH   9
@@ -51,18 +53,19 @@ static CompMetadata winrulesMetadata;
 static int displayPrivateIndex;
 
 typedef struct _WinrulesWindow {
-
     unsigned int allowedActions;
     unsigned int stateSetMask;
     unsigned int protocolSetMask;
-    Bool widgetSet;
 
-    Bool firstMap;
+    Bool hasAlpha;
 } WinrulesWindow;
 
 typedef struct _WinrulesDisplay {
     int screenPrivateIndex;
-    HandleEventProc handleEvent;
+
+    HandleEventProc            handleEvent;
+    MatchExpHandlerChangedProc matchExpHandlerChanged;
+    MatchPropertyChangedProc   matchPropertyChanged;
 } WinrulesDisplay;
 
 typedef struct _WinrulesScreen {
@@ -71,75 +74,74 @@ typedef struct _WinrulesScreen {
     CompOption opt[WINRULES_SCREEN_OPTION_NUM];
 } WinrulesScreen;
 
-#define GET_WINRULES_DISPLAY(d)				     			   \
-    ((WinrulesDisplay *) (d)->privates[displayPrivateIndex].ptr)
+#define GET_WINRULES_DISPLAY(d)				\
+    ((WinrulesDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
-#define WINRULES_DISPLAY(d)			   				   \
+#define WINRULES_DISPLAY(d)			   	\
     WinrulesDisplay *wd = GET_WINRULES_DISPLAY (d)
 
-#define GET_WINRULES_SCREEN(s, wd)					 	   \
-    ((WinrulesScreen *) (s)->privates[(wd)->screenPrivateIndex].ptr)
+#define GET_WINRULES_SCREEN(s, wd)				    \
+    ((WinrulesScreen *) (s)->base.privates[(wd)->screenPrivateIndex].ptr)
 
-#define WINRULES_SCREEN(s)							   \
-    WinrulesScreen *ws = GET_WINRULES_SCREEN (s, GET_WINRULES_DISPLAY (s->display))
+#define WINRULES_SCREEN(s)			    \
+    WinrulesScreen *ws = GET_WINRULES_SCREEN (s,    \
+			 GET_WINRULES_DISPLAY (s->display))
 
-#define GET_WINRULES_WINDOW(w, ws)                                   		   \
-    ((WinrulesWindow *) (w)->privates[(ws)->windowPrivateIndex].ptr)
-    
-#define WINRULES_WINDOW(w)							   \
-    WinrulesWindow *ww = GET_WINRULES_WINDOW  (w,				   \
-					    GET_WINRULES_SCREEN  (w->screen,	   \
-					    GET_WINRULES_DISPLAY (w->screen->display)))
+#define GET_WINRULES_WINDOW(w, ws)                                  \
+    ((WinrulesWindow *) (w)->base.privates[(ws)->windowPrivateIndex].ptr)
+
+#define WINRULES_WINDOW(w)					\
+    WinrulesWindow *ww = GET_WINRULES_WINDOW  (w,		\
+   			 GET_WINRULES_SCREEN  (w->screen,	\
+			 GET_WINRULES_DISPLAY (w->screen->display)))
 
 #define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))
 
-
 static void
-winrulesSetProtocols (CompDisplay *display,
+winrulesSetProtocols (CompDisplay  *display,
 		      unsigned int protocols,
-		      Window id)
+		      Window       id)
 {
-    Atom *protocol = NULL;
-    int count = 0;
+    Atom protocol[4];
+    int  count = 0;
 
     if (protocols & CompWindowProtocolDeleteMask)
-    {
-	protocol = realloc (protocol, (count + 1) * sizeof(Atom));
 	protocol[count++] = display->wmDeleteWindowAtom;
-    }
     if (protocols & CompWindowProtocolTakeFocusMask)
-    {
-	protocol = realloc (protocol, (count + 1) * sizeof(Atom));
 	protocol[count++] = display->wmTakeFocusAtom;
-    }
     if (protocols & CompWindowProtocolPingMask)
-    {
-	protocol = realloc (protocol, (count + 1) * sizeof(Atom));
 	protocol[count++] = display->wmPingAtom;
-    }
     if (protocols & CompWindowProtocolSyncRequestMask)
-    {
-	protocol = realloc (protocol, (count + 1) * sizeof(Atom));
-	protocol[count] = display->wmSyncRequestAtom;
-    }
-    XSetWMProtocols (display->display,
-		     id,
-		     protocol,
-		     count);
+	protocol[count++] = display->wmSyncRequestAtom;
 
-    XFree (protocol);
+    XSetWMProtocols (display->display, id, protocol, count);
+}
+
+static Bool
+isWinrulesWindow (CompWindow *w)
+{
+    if (w->attrib.override_redirect)
+	return FALSE;
+
+    if (!(w->type & WINRULES_TARGET_WINDOWS))
+	return FALSE;
+
+    return TRUE;
 }
 
 /* FIXME? Directly set inputHint, not a problem for now */
 static void
 winrulesSetNoFocus (CompWindow *w,
-		    int optNum)
+		    int        optNum)
 {
     unsigned int newProtocol = w->protocols;
 
     WINRULES_SCREEN (w->screen);
     WINRULES_WINDOW (w);
 
+    if (!isWinrulesWindow (w))
+	return;
+
     if (matchEval (&ws->opt[optNum].value.match, w))
     {
 	if (w->protocols & CompWindowProtocolTakeFocusMask)
@@ -162,16 +164,40 @@ winrulesSetNoFocus (CompWindow *w,
 		      w->id);
 }
 
+static void
+winrulesSetNoAlpha (CompWindow *w,
+	  	    int        optNum)
+{
+    WINRULES_SCREEN (w->screen);
+    WINRULES_WINDOW (w);
+
+    if (!isWinrulesWindow (w))
+	return;
+
+    if (matchEval (&ws->opt[optNum].value.match, w))
+    {
+	ww->hasAlpha = w->alpha;
+	w->alpha = FALSE;
+    }
+    else
+    {
+	w->alpha = ww->hasAlpha;
+    }
+}
+
 static void
 winrulesUpdateState (CompWindow *w,
-		     int optNum,
-		     int mask)
+		     int        optNum,
+		     int        mask)
 {
     unsigned int newState = w->state;
 
     WINRULES_SCREEN (w->screen);
     WINRULES_WINDOW (w);
 
+    if (!isWinrulesWindow (w))
+	return;
+
     if (matchEval (&ws->opt[optNum].value.match, w))
     {
 	newState |= mask;
@@ -188,9 +214,6 @@ winrulesUpdateState (CompWindow *w,
     {
 	changeWindowState (w, newState);
 
-	recalcWindowType (w);
-	recalcWindowActions (w);
-
 	if (mask & (CompWindowStateFullscreenMask |
 		    CompWindowStateAboveMask      |
 		    CompWindowStateBelowMask       ))
@@ -200,48 +223,20 @@ winrulesUpdateState (CompWindow *w,
     }
 }
 
-static void
-winrulesUpdateWidget (CompWindow *w)
-{
-    Atom compizWidget = XInternAtom (w->screen->display->display,
-				     "_COMPIZ_WIDGET",
-				     FALSE);
-
-    WINRULES_SCREEN (w->screen);
-    WINRULES_WINDOW (w);
-
-    if (matchEval
-	(&ws->opt[WINRULES_SCREEN_OPTION_WIDGET_MATCH].value.match, w))
-    {
-	if (w->inShowDesktopMode || w->mapNum ||
-                w->attrib.map_state == IsViewable || w->minimized)
-	{
-	    if (w->minimized || w->inShowDesktopMode)
-		unminimizeWindow (w);
-	    XChangeProperty (w->screen->display->display, w->id, compizWidget,
-			     XA_STRING, 8, PropModeReplace,
-			     (unsigned char *)(int[]){-2}, 1);
-	    ww->widgetSet = TRUE;
-	}
-    }
-    else if (ww->widgetSet)
-    {
-	XDeleteProperty (w->screen->display->display, w->id, compizWidget);
-	ww->widgetSet = FALSE;
-    }
-}
-
 static void
 winrulesSetAllowedActions (CompWindow *w,
-			   int optNum,
-			   int action)
+			   int        optNum,
+			   int        action)
 {
     WINRULES_SCREEN (w->screen);
     WINRULES_WINDOW (w);
 
+    if (!isWinrulesWindow (w))
+	return;
+
     if (matchEval (&ws->opt[optNum].value.match, w))
 	ww->allowedActions &= ~action;
-    else if ( !(ww->allowedActions & action))
+    else if (!(ww->allowedActions & action))
 	ww->allowedActions |= action;
 
     recalcWindowActions (w);
@@ -257,6 +252,9 @@ winrulesMatchSizeValue (CompWindow *w,
 {
     int i, min;
 
+    if (!isWinrulesWindow (w))
+	return FALSE;
+
     if (w->type & CompWindowTypeDesktopMask)
 	return FALSE;
 
@@ -285,31 +283,31 @@ winrulesMatchSize (CompWindow *w,
     WINRULES_SCREEN (w->screen);
 
     return winrulesMatchSizeValue (w,
-				   &ws->opt[WINRULES_SCREEN_OPTION_SIZE_MATCHES],
-				   &ws->opt[WINRULES_SCREEN_OPTION_SIZE_WIDTH_VALUES],
-				   &ws->opt[WINRULES_SCREEN_OPTION_SIZE_HEIGHT_VALUES],
-				   width,
-				   height);
+	&ws->opt[WINRULES_SCREEN_OPTION_SIZE_MATCHES],
+	&ws->opt[WINRULES_SCREEN_OPTION_SIZE_WIDTH_VALUES],
+	&ws->opt[WINRULES_SCREEN_OPTION_SIZE_HEIGHT_VALUES],
+	width, height);
 }
 
 static void
 winrulesUpdateWindowSize (CompWindow *w,
-			  int width,
-			  int height)
+			  int        width,
+			  int        height)
 {
     XWindowChanges xwc;
-    unsigned int xwcm = 0;
+    unsigned int   xwcm = 0;
 
     if (width != w->serverWidth)
 	xwcm |= CWWidth;
     if (height != w->serverHeight)
 	xwcm |= CWHeight;
 
-    xwc.x = w->serverX;
-    xwc.y = w->serverY;
     xwc.width = width;
     xwc.height = height;
 
+    if (w->mapNum && xwcm)
+	sendSyncRequest (w);
+
     configureXWindow (w, xwcm, &xwc);
 }
 
@@ -327,12 +325,12 @@ winrulesGetScreenOptions (CompPlugin *plugin,
 static Bool
 winrulesSetScreenOption (CompPlugin *plugin,
 			 CompScreen      *screen,
-                         char            *name, 
+                         const char      *name,
                          CompOptionValue *value)
 {
-    CompOption *o;
-    CompWindow *w;
-    int index;
+    CompOption   *o;
+    int          index;
+    unsigned int updateStateMask = 0, updateActionsMask = 0;
 
     WINRULES_SCREEN (screen);
 
@@ -344,204 +342,78 @@ winrulesSetScreenOption (CompPlugin *plugin,
     {
     case WINRULES_SCREEN_OPTION_SKIPTASKBAR_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_SKIPTASKBAR_MATCH,
-				     CompWindowStateSkipTaskbarMask);
-	    }
-				
-	    return TRUE;
-	}
+	    updateStateMask = CompWindowStateSkipTaskbarMask;
 	break;
 
     case WINRULES_SCREEN_OPTION_SKIPPAGER_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_SKIPPAGER_MATCH,
-				     CompWindowStateSkipPagerMask);
-	    }
-	    return TRUE;
-	}
+	    updateStateMask = CompWindowStateSkipPagerMask;
 	break;
 
     case WINRULES_SCREEN_OPTION_ABOVE_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_ABOVE_MATCH,
-				     CompWindowStateAboveMask);
-	    }
-	    return TRUE;
-	}
+	    updateStateMask = CompWindowStateAboveMask;
 	break;
 	
     case WINRULES_SCREEN_OPTION_BELOW_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_BELOW_MATCH,
-				     CompWindowStateBelowMask);
-	    }
-	    return TRUE;
-	}
+	    updateStateMask = CompWindowStateBelowMask;
 	break;
 	
     case WINRULES_SCREEN_OPTION_STICKY_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_STICKY_MATCH,
-				     CompWindowStateStickyMask);
-	    }
-	    return TRUE;
-	}
+	    updateStateMask = CompWindowStateStickyMask;
 	break;
 	
     case WINRULES_SCREEN_OPTION_FULLSCREEN_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_FULLSCREEN_MATCH,
-				     CompWindowStateFullscreenMask);
-	    }
-	    return TRUE;
-	}
+	    updateStateMask = CompWindowStateFullscreenMask;
 	break;
 
     case WINRULES_SCREEN_OPTION_NOMOVE_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesSetAllowedActions (w,
-					   WINRULES_SCREEN_OPTION_NOMOVE_MATCH,
-					   CompWindowActionMoveMask);
-	    }
-	    return TRUE;
-	}
+	    updateActionsMask = CompWindowActionMoveMask;
 	break;
 
     case WINRULES_SCREEN_OPTION_NORESIZE_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesSetAllowedActions (w,
-					   WINRULES_SCREEN_OPTION_NORESIZE_MATCH,
-					   CompWindowActionResizeMask);
-	    }
-	    return TRUE;
-	}
+	    updateActionsMask = CompWindowActionResizeMask;
 	break;
 
     case WINRULES_SCREEN_OPTION_NOMINIMIZE_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesSetAllowedActions (w, 
-					   WINRULES_SCREEN_OPTION_NOMINIMIZE_MATCH,
-					   CompWindowActionMinimizeMask);
-	    }
-	    return TRUE;
-	}
+	    updateActionsMask = CompWindowActionMinimizeMask;
 	break;
 
     case WINRULES_SCREEN_OPTION_NOMAXIMIZE_MATCH:
 	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-
-		winrulesSetAllowedActions (w, 
-					   WINRULES_SCREEN_OPTION_NOMAXIMIZE_MATCH,
-					   CompWindowActionMaximizeVertMask|
-					   CompWindowActionMaximizeHorzMask);
-	    }
-	    return TRUE;
-	}
+	    updateActionsMask = CompWindowActionMaximizeVertMask |
+		                CompWindowActionMaximizeHorzMask;
 	break;
 
     case WINRULES_SCREEN_OPTION_NOCLOSE_MATCH:
+	if (compSetMatchOption (o, value))
+	    updateActionsMask = CompWindowActionCloseMask;
+	break;
+    case WINRULES_SCREEN_OPTION_NOARGB_MATCH:
 	if (compSetMatchOption (o, value))
 	{
+	    CompWindow *w;
+
 	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
+		winrulesSetNoAlpha (w, WINRULES_SCREEN_OPTION_NOARGB_MATCH);
 
-		winrulesSetAllowedActions (w, 
-					   WINRULES_SCREEN_OPTION_NOCLOSE_MATCH,
-					   CompWindowActionCloseMask);
-	    }
 	    return TRUE;
 	}
 	break;
     case WINRULES_SCREEN_OPTION_NOFOCUS_MATCH:
 	if (compSetMatchOption (o, value))
 	{
+	    CompWindow *w;
+
 	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
-		
 		winrulesSetNoFocus (w, WINRULES_SCREEN_OPTION_NOFOCUS_MATCH);
-	    }
-	    return TRUE;
-	}
-	break;
-    case WINRULES_SCREEN_OPTION_WIDGET_MATCH:
-	if (compSetMatchOption (o, value))
-	{
-	    for (w = screen->windows; w; w = w->next)
-	    {
-		if (!w->type & WINRULES_TARGET_WINDOWS)
-		    continue;
 
-	    	winrulesUpdateWidget (w);
-	    }
 	    return TRUE;
 	}
 	break;
@@ -562,6 +434,26 @@ winrulesSetScreenOption (CompPlugin *plugin,
         break;
     }
 
+    if (updateStateMask)
+    {
+	CompWindow *w;
+
+	for (w = screen->windows; w; w = w->next)
+	    winrulesUpdateState (w, index, updateStateMask);
+
+	return TRUE;
+    }
+
+    if (updateActionsMask)
+    {
+	CompWindow *w;
+
+	for (w = screen->windows; w; w = w->next)
+	    winrulesSetAllowedActions (w, index, updateActionsMask);
+
+	return TRUE;
+    }
+
     return FALSE;
 }
 
@@ -573,78 +465,11 @@ winrulesHandleEvent (CompDisplay *d,
 
     WINRULES_DISPLAY (d);
 
-    if (event->type == MapNotify)
+    if (event->type == MapRequest)
     {
 	w = findWindowAtDisplay (d, event->xmap.window);
-	if (w && w->type & WINRULES_TARGET_WINDOWS)
-	{
-	    WINRULES_WINDOW (w);
-	    /* Only apply at window creation.
-	     * Using CreateNotify not working.
-	     */
-	    if (ww->firstMap)
-	    {
-		int width, height;
-		
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_SKIPTASKBAR_MATCH,
-				     CompWindowStateSkipTaskbarMask);
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_SKIPPAGER_MATCH,
-				     CompWindowStateSkipPagerMask);
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_ABOVE_MATCH,
-				     CompWindowStateAboveMask);
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_BELOW_MATCH,
-				     CompWindowStateBelowMask);
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_STICKY_MATCH,
-				     CompWindowStateStickyMask);
-
-		winrulesUpdateState (w,
-				     WINRULES_SCREEN_OPTION_FULLSCREEN_MATCH,
-				     CompWindowStateFullscreenMask);
-
-		winrulesUpdateWidget (w);
-
-		winrulesSetAllowedActions (w,
-					   WINRULES_SCREEN_OPTION_NOMOVE_MATCH,
-					   CompWindowActionMoveMask);
-
-		winrulesSetAllowedActions (w,
-					   WINRULES_SCREEN_OPTION_NORESIZE_MATCH,
-					   CompWindowActionResizeMask);
-
-		winrulesSetAllowedActions (w,
-					   WINRULES_SCREEN_OPTION_NOMINIMIZE_MATCH,
-					   CompWindowActionMinimizeMask);
-
-		winrulesSetAllowedActions (w,
-					   WINRULES_SCREEN_OPTION_NOMAXIMIZE_MATCH,
-					   CompWindowActionMaximizeVertMask|
-					   CompWindowActionMaximizeHorzMask);
-	
-		winrulesSetAllowedActions (w,
-					   WINRULES_SCREEN_OPTION_NOCLOSE_MATCH,
-					   CompWindowActionCloseMask);
-		
-		if (winrulesMatchSize (w, &width, &height))
-		    winrulesUpdateWindowSize (w, width, height);
-	    }
-
-	    ww->firstMap = FALSE;
-	}
-    }
-    else if (event->type == MapRequest)
-    {
-	w = findWindowAtDisplay (d, event->xmap.window);
-	if (w && w->type & WINRULES_TARGET_WINDOWS)
-	    winrulesSetNoFocus (w,WINRULES_SCREEN_OPTION_NOFOCUS_MATCH);
+	if (w)
+	    winrulesSetNoFocus (w, WINRULES_SCREEN_OPTION_NOFOCUS_MATCH);
     }
 
     UNWRAP (wd, d, handleEvent);
@@ -669,11 +494,105 @@ winrulesGetAllowedActionsForWindow (CompWindow   *w,
 }
 
 static Bool
-winrulesInitDisplay (CompPlugin  *p, 
+winrulesApplyRules (void *closure)
+{
+    CompWindow *w = (CompWindow *) closure;
+    int        width, height;
+
+    winrulesUpdateState (w,
+			 WINRULES_SCREEN_OPTION_SKIPTASKBAR_MATCH,
+			 CompWindowStateSkipTaskbarMask);
+
+    winrulesUpdateState (w,
+			 WINRULES_SCREEN_OPTION_SKIPPAGER_MATCH,
+			 CompWindowStateSkipPagerMask);
+
+    winrulesUpdateState (w,
+			 WINRULES_SCREEN_OPTION_ABOVE_MATCH,
+			 CompWindowStateAboveMask);
+
+    winrulesUpdateState (w,
+			 WINRULES_SCREEN_OPTION_BELOW_MATCH,
+			 CompWindowStateBelowMask);
+
+    winrulesUpdateState (w,
+			 WINRULES_SCREEN_OPTION_STICKY_MATCH,
+			 CompWindowStateStickyMask);
+
+    winrulesUpdateState (w,
+			 WINRULES_SCREEN_OPTION_FULLSCREEN_MATCH,
+			 CompWindowStateFullscreenMask);
+
+    winrulesSetAllowedActions (w,
+			       WINRULES_SCREEN_OPTION_NOMOVE_MATCH,
+			       CompWindowActionMoveMask);
+
+    winrulesSetAllowedActions (w,
+			       WINRULES_SCREEN_OPTION_NORESIZE_MATCH,
+			       CompWindowActionResizeMask);
+
+    winrulesSetAllowedActions (w,
+			       WINRULES_SCREEN_OPTION_NOMINIMIZE_MATCH,
+			       CompWindowActionMinimizeMask);
+
+    winrulesSetAllowedActions (w,
+			       WINRULES_SCREEN_OPTION_NOMAXIMIZE_MATCH,
+			       CompWindowActionMaximizeVertMask |
+			       CompWindowActionMaximizeHorzMask);
+
+    winrulesSetAllowedActions (w,
+			       WINRULES_SCREEN_OPTION_NOCLOSE_MATCH,
+			       CompWindowActionCloseMask);
+
+    winrulesSetNoAlpha (w, WINRULES_SCREEN_OPTION_NOARGB_MATCH);
+
+    if (winrulesMatchSize (w, &width, &height))
+	winrulesUpdateWindowSize (w, width, height);
+
+    return FALSE;
+}
+
+
+static void
+winrulesMatchExpHandlerChanged (CompDisplay *d)
+{
+    CompScreen *s;
+    CompWindow *w;
+
+    WINRULES_DISPLAY (d);
+
+    UNWRAP (wd, d, matchExpHandlerChanged);
+    (*d->matchExpHandlerChanged) (d);
+    WRAP (wd, d, matchExpHandlerChanged, winrulesMatchExpHandlerChanged);
+
+    /* match options are up to date after the call to matchExpHandlerChanged */
+    for (s = d->screens; s; s = s->next)
+	for (w = s->windows; w; w = w->next)
+	    winrulesApplyRules (w);
+}
+
+static void
+winrulesMatchPropertyChanged (CompDisplay *d,
+	    		      CompWindow  *w)
+{
+    WINRULES_DISPLAY (d);
+
+    winrulesApplyRules (w);
+
+    UNWRAP (wd, d, matchPropertyChanged);
+    (*d->matchPropertyChanged) (d, w);
+    WRAP (wd, d, matchPropertyChanged, winrulesMatchPropertyChanged);
+}
+
+static Bool
+winrulesInitDisplay (CompPlugin  *p,
 		     CompDisplay *d)
 {
     WinrulesDisplay *wd;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     wd = malloc (sizeof (WinrulesDisplay));
     if (!wd)
         return FALSE;
@@ -684,20 +603,28 @@ winrulesInitDisplay (CompPlugin  *p,
         free (wd);
         return FALSE;
     }
+
     WRAP (wd, d, handleEvent, winrulesHandleEvent);
-    d->privates[displayPrivateIndex].ptr = wd;
+    WRAP (wd, d, matchExpHandlerChanged, winrulesMatchExpHandlerChanged);
+    WRAP (wd, d, matchPropertyChanged, winrulesMatchPropertyChanged);
+
+    d->base.privates[displayPrivateIndex].ptr = wd;
 
     return TRUE;
 }
 
 static void
-winrulesFiniDisplay (CompPlugin  *p, 
+winrulesFiniDisplay (CompPlugin  *p,
 		     CompDisplay *d)
 {
     WINRULES_DISPLAY (d);
 
     freeScreenPrivateIndex (d, wd->screenPrivateIndex);
+
     UNWRAP (wd, d, handleEvent);
+    UNWRAP (wd, d, matchExpHandlerChanged);
+    UNWRAP (wd, d, matchPropertyChanged);
+
     free (wd);
 }
 
@@ -708,7 +635,7 @@ static const CompMetadataOptionInfo winrulesScreenOptionInfo[] = {
     { "below_match", "match", 0, 0, 0 },
     { "sticky_match", "match", 0, 0, 0 },
     { "fullscreen_match", "match", 0, 0, 0 },
-    { "widget_match", "match", 0, 0, 0 },
+    { "no_argb_match", "match", 0, 0, 0 },
     { "no_move_match", "match", 0, 0, 0 },
     { "no_resize_match", "match", 0, 0, 0 },
     { "no_minimize_match", "match", 0, 0, 0 },
@@ -721,7 +648,7 @@ static const CompMetadataOptionInfo winrulesScreenOptionInfo[] = {
 };
 
 static Bool
-winrulesInitScreen (CompPlugin *p, 
+winrulesInitScreen (CompPlugin *p,
 		    CompScreen *s)
 {
     WinrulesScreen *ws;
@@ -753,13 +680,13 @@ winrulesInitScreen (CompPlugin *p,
     WRAP (ws, s, getAllowedActionsForWindow,
 	  winrulesGetAllowedActionsForWindow);
 
-    s->privates[wd->screenPrivateIndex].ptr = ws;
+    s->base.privates[wd->screenPrivateIndex].ptr = ws;
 
     return TRUE;
 }
 
 static void
-winrulesFiniScreen (CompPlugin *p, 
+winrulesFiniScreen (CompPlugin *p,
                     CompScreen *s)
 {
     WINRULES_SCREEN (s);
@@ -774,32 +701,33 @@ winrulesFiniScreen (CompPlugin *p,
 }
 
 static Bool
-winrulesInitWindow (CompPlugin *p, 
+winrulesInitWindow (CompPlugin *p,
 		    CompWindow *w)
 {
+    WinrulesWindow *ww;
+
     WINRULES_SCREEN (w->screen);
 
-    WinrulesWindow *ww = malloc (sizeof (WinrulesWindow));
+    ww = malloc (sizeof (WinrulesWindow));
     if (!ww)
-    {
         return FALSE;
-    }
 
-    ww->widgetSet       = FALSE;
     ww->stateSetMask    = 0;
     ww->protocolSetMask = 0;
 
     ww->allowedActions = ~0;
 
-    ww->firstMap = TRUE;
+    ww->hasAlpha = w->alpha;
 
-    w->privates[ws->windowPrivateIndex].ptr = ww;
+    w->base.privates[ws->windowPrivateIndex].ptr = ww;
+
+    compAddTimeout (0, winrulesApplyRules, w);
 
     return TRUE;
 }
 
 static void
-winrulesFiniWindow (CompPlugin *p, 
+winrulesFiniWindow (CompPlugin *p,
                     CompWindow *w)
 {
     WINRULES_WINDOW (w);
@@ -807,6 +735,65 @@ winrulesFiniWindow (CompPlugin *p,
     free (ww);
 }
 
+static CompBool
+winrulesInitObject (CompPlugin *p,
+		    CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) winrulesInitDisplay,
+	(InitPluginObjectProc) winrulesInitScreen,
+	(InitPluginObjectProc) winrulesInitWindow
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+winrulesFiniObject (CompPlugin *p,
+		    CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) winrulesFiniDisplay,
+	(FiniPluginObjectProc) winrulesFiniScreen,
+	(FiniPluginObjectProc) winrulesFiniWindow
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
+static CompOption *
+winrulesGetObjectOptions (CompPlugin *plugin,
+	  		  CompObject *object,
+	  		  int	     *count)
+{
+    static GetPluginObjectOptionsProc dispTab[] = {
+	(GetPluginObjectOptionsProc) 0, /* GetCoreOptions */
+	(GetPluginObjectOptionsProc) 0, /* GetDisplayOptions */
+	(GetPluginObjectOptionsProc) winrulesGetScreenOptions
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
+		     (void *) (*count = 0), (plugin, object, count));
+}
+
+static CompBool
+winrulesSetObjectOption (CompPlugin      *plugin,
+	  		 CompObject      *object,
+	  		 const char      *name,
+	  		 CompOptionValue *value)
+{
+    static SetPluginObjectOptionProc dispTab[] = {
+	(SetPluginObjectOptionProc) 0, /* SetCoreOption */
+	(SetPluginObjectOptionProc) 0, /* SetDisplayOption */
+	(SetPluginObjectOptionProc) winrulesSetScreenOption
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab), FALSE,
+		     (plugin, object, name, value));
+}
+
 static Bool
 winrulesInit (CompPlugin *p)
 {
@@ -837,13 +824,6 @@ winrulesFini (CompPlugin *p)
     compFiniMetadata (&winrulesMetadata);
 }
 
-static int
-winrulesGetVersion (CompPlugin *plugin,
-		int	    version)
-{
-    return ABIVERSION;
-}
-
 static CompMetadata *
 winrulesGetMetadata (CompPlugin *plugin)
 {
@@ -852,24 +832,17 @@ winrulesGetMetadata (CompPlugin *plugin)
 
 static CompPluginVTable winrulesVTable = {
     "winrules",
-    winrulesGetVersion,
     winrulesGetMetadata,
     winrulesInit,
     winrulesFini,
-    winrulesInitDisplay,
-    winrulesFiniDisplay,
-    winrulesInitScreen,
-    winrulesFiniScreen,
-    winrulesInitWindow,
-    winrulesFiniWindow,
-    0, /* winrulesGetDisplayOptions, */
-    0, /* winrulesSetDisplayOption,  */
-    winrulesGetScreenOptions,
-    winrulesSetScreenOption
+    winrulesInitObject,
+    winrulesFiniObject,
+    winrulesGetObjectOptions,
+    winrulesSetObjectOption,
 };
 
 CompPluginVTable *
-getCompPluginInfo (void)
+getCompPluginInfo20070830 (void)
 {
     return &winrulesVTable;
 }
